# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#
from htf.daqmx.core import DAQError
from htf.daqmx.task import Task
from ctypes import *
from htf.daqmx.api import int32,\
    float64, DAQmx_Val_GroupByChannel, DAQmx_Val_FiniteSamps, DAQmx_Val_Rising,\
    DAQmx_Val_Volts, DAQmx_Val_DoNotAllowRegen,\
    DAQmx_Val_OnBrdMemHalfFullOrLess, DAQmx_Val_ContSamps
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageOutputPort(DelegateMixin):
    """
    ``AnalogVoltageOutputPort`` is used to access a single analog output port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(self, descriptor,
                 minimum_voltage=-10.0,
                 maximum_voltage=10.0,
                 number_of_lines=32,
                 number_of_samples=2,
                 sample_frequency=1000.0,
                 timeout=1.0,
                 continuous_mode=False):
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`
            minimum_voltage=-10.0 (float): the minimum voltage.
            maximum_voltage=10.0 (float): the maximum voltage.
            number_of_lines=32 (int): the number of lines defined in the
                descriptor. For example the descriptor "cDaq1Mod3/ai0:31"
                defines 32 lines.
            number_of_samples=2 (int): the number of samples to write for every
                line.
            sample_frequency=1000.0 (float): the sample frequency in Hertz.
            timeout=1.0 (float): the timeout when acquiring a new value in
                seconds.
            continuous_mode=False: if set to ``True`` continuous sampling is
                done and get() becomes blocking. Simply read samples in a loop.
        """

        task_name = "AO_" + descriptor.replace("/", "_")
        self.number_of_samples = number_of_samples
        self.sample_frequency = sample_frequency
        self.number_of_lines = number_of_lines
        self.timeout = timeout
        self.continuous_mode = continuous_mode
        self.data = cast(
            (float64 * self.number_of_samples * self.number_of_lines)(),
            POINTER(float64))
        self.values = None
        self.written = int32(0)
        self.task = Task(task_name)

        if not isinstance(descriptor, bytes):
            descriptor = descriptor.encode()

        self.task.CreateAOVoltageChan(descriptor,
                                      None,
                                      float64(minimum_voltage),
                                      float64(maximum_voltage),
                                      DAQmx_Val_Volts,
                                      None)

        if self.continuous_mode is False:
            self.task.CfgSampClkTiming(None,
                                       float(self.sample_frequency),
                                       DAQmx_Val_Rising,
                                       DAQmx_Val_FiniteSamps,
                                       self.number_of_samples * self.number_of_lines)
        else:
            self.task.CfgSampClkTiming(None,
                                       float(self.sample_frequency),
                                       DAQmx_Val_Rising, DAQmx_Val_ContSamps,
                                       self.number_of_samples * self.number_of_lines)
            self.task.SetWriteRegenMode(DAQmx_Val_DoNotAllowRegen)
            self.task.SetAODataXferReqCond(descriptor,
                                           DAQmx_Val_OnBrdMemHalfFullOrLess)
            self.task.SetBufOutputBufSize(2 * int(self.number_of_samples))

    def set(self, values):
        """
        Write analog data.

        Args:
            values: samples to be written on the output port.

        ``values`` are expected to have the following layout::

            [
                [line0 sample0, line0 sample1, ..., line0 sampleN-1],
                [line1 sample0, line1 sample1, ..., line1 sampleN-1],
                ...
                [lineM-1 sample0, lineM-1 sample1, ..., lineM-1 sampleN-1]
            ]
        """

        self.values = values

        for line in range(self.number_of_lines):
            for sample in range(self.number_of_samples):
                self.data[line * self.number_of_samples + sample] = \
                    values[line][sample]

        self.task.WriteAnalogF64(self.number_of_samples,
                                 False,
                                 float64(self.timeout),
                                 DAQmx_Val_GroupByChannel,
                                 self.data,
                                 byref(self.written),
                                 None)
        self.task.start()
        self.task.WaitUntilTaskDone(self.timeout)
        self.task.stop()

    def get(self):
        """
        Return the last written analog values.

        Returned data layout::

            [
                [line0 sample0, line0 sample1, ..., line0 sampleN-1],
                [line1 sample0, line1 sample1, ..., line1 sampleN-1],
                ...
                [lineM-1 sample0, lineM-1 sample1, ..., lineM-1 sampleN-1]
            ]
        """
        return self.values

    def close(self):
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self):
        self.close()

    def __str__(self):
        return str(self.get())


if __name__ == "__main__":
    ao0 = AnalogVoltageOutputPort("Dev1/ao0", number_of_lines=2)

    read = ao0.get()
    print("read:", read)
    data = [[1.0, 2.0], [3.0, 4.0]]
    ao0.set(data)
    read = ao0.get()
    print("read:", read)
